import networkx as nx
import numpy as np

from copy import deepcopy
from collections import defaultdict
from ylearn.utils import to_repr
from . import prob
from .utils import (check_nodes, ancestors_of_iter, descendents_of_iter)


class CausalGraph:
    """
    A class for representing DAGs of causal structures.

    Attributes
    ----------
    causation : dict
        Descriptions of the causal structures where values are parents of the
        corresponding keys.
    
    dag : nx.MultiDiGraph
        Graph represented by the networkx package.
    
    prob : ylearn.causal_model.prob.Prob
        The encoded probability distribution of the causal graph.
    
    latent_confounding_arcs : list of tuple of two str
        Two elements in the tuple are names of nodes in the graph where there
        exists an latent confounding arcs between them. Semi-Markovian graphs
        with unobserved confounders can be converted to a graph without
        unobserved variables, where one can add bi-directed latent confounding
        arcs represent these relations. For example, the causal graph X <- U -> Y,
        where U is an unobserved confounder of X and Y, can be converted
        equivalently to X <-->Y where <--> is a latent confounding arc.
    
    is_dag : bool
        Determine whether the graph is a DAG, which is a necessary condition 
        for it to be a valid causal graph.
    
    c_components : set
        The C-components of the graph.
    
    observed_dag : nx.MultiDiGraph
        A causal graph with only observed variables.
    
    topo_order : list
        The topological order of the graph.
    
    explicit_unob_var_dag : nx.MultiDiGraph
        A new dag where all unobserved confounding arcs are replaced
        by explicit unobserved variables. See latent_confounding_arcs for more 
        details of the unobserved variables.

    Methods
    ----------
    to_adj_matrix()
        Return the numpy matrix of the adjecency matrix.
    
    to_adj_list()
        Return the numpy array of the adjecency matrix.
    
    ancestors(y)
        Return ancestors of y.
    
    add_nodes(nodes, new=False)
        If not new, add all nodes in the nodes to the current
        CausalGraph, else create a new graph and add nodes.
    
    add_edges_from(edge_list, new=False, observed=True)
        Add all edges in the edge_list to the CausalGraph.
    
    parents(x, observed=True)
        Find the parents of the node x in the CausalGraph.
    
    add_edge(i, j, observed=True)
        Add an edge between nodes i and j to the CausalGraph. Add an unobserved
        confounding arc if not observed.
    
    remove_nodes(nodes, new=False)
        Remove all nodes in the graph. If new, do this in a new CausalGraph.
    
    remove_edge(i, j, observed=True)
        Remove the edge in the CausalGraph. If observed, remove the unobserved
        latent confounding arcs.
    
    remove_edges_from(edge_list, new=False, observed=True)
        Remove all edges in the edge_list in the CausalGraph.
    
    build_sub_graph(subset)
        Return a new CausalGraph as the subgraph of self with nodes in the
        subset.
    
    remove_incoming_edges(y, new=False)
        Remove all incoming edges of all nodes in y. If new, return a new
        CausalGraph.
    
    remove_outgoing_edges(y, new=False)
        Remove all outgoing edges of all nodes in y. If new, return a new
        CausalGraph.
    """

    def __init__(self, causation, dag=None, latent_confounding_arcs=None):
        """
        Parameters
        ----------
        causation : dict
            Descriptions of the causal structures where values are parents of the
            corresponding keys.        
        dag : nx.MultiGraph, optional
            A konw graph structure represented. If provided, dag must represent
            the causal structures stored in causation. Defaults to None.
        
        latent_confounding_arcs : set or list of tuple of two str, optional
            Two elements in the tuple are names of nodes in the graph where there
            exists an latent confounding arcs between them. Semi-Markovian graphs
            with unobserved confounders can be converted to a graph without
            unobserved variables, where one can add bi-directed latent confounding
            arcs to represent these relations. For example, the causal graph X <- U -> Y,
            where U is an unobserved confounder of X and Y, can be converted
            equivalently to X <-->Y where <--> is a latent confounding arc.
        """
        self.causation = defaultdict(list, causation)
        self.ava_nodes = self.causation.keys()
        self.dag = self.observed_dag.copy() if dag is None else dag

        # add unobserved bidirected confounding arcs to the graph, with the
        # letter 'n' representing that the edge is unobserved
        if latent_confounding_arcs is not None:
            for edge in latent_confounding_arcs:
                self.dag.add_edges_from(
                    [(edge[0], edge[1], 'n'), (edge[1], edge[0], 'n')]
                )
    
    @property
    def prob(self):
        """The encoded probability distribution.

        Returns
        ----------
        Prob
        """
        return prob.Prob(variables=set(self.causation.keys()))

    @property
    def latent_confounding_arcs(self):
        """Return the latent confounding arcs encoded in the graph.

        Returns
        ----------
        list

        """
        W = nx.to_numpy_matrix(self.dag)
        a, w_t = np.where(W >= 1), W.T.A
        arcs, nodes = [], list(self.dag.nodes)
        for row, col in zip(a[0], a[1]):
            if w_t[row][col] >= 1 and (nodes[col], nodes[row]) not in arcs:
                arcs.append((nodes[row], nodes[col]))
        return arcs

    @property
    def is_dag(self):
        """Verify whether the constructed graph is a DAG.
        """
        # TODO: determin if the graph is a DAG, try tr(e^{W\circledot W}-d)=0
        return nx.is_directed_acyclic_graph(self.observed_dag)

    def to_adj_matrix(self):
        """Return the adjacency matrix.
        """
        W = nx.to_numpy_matrix(self.dag)
        return W

    # def to_adj_list(self):
    #     """Return the adjacency list."""
    #     pass

    def is_d_separated(self, x, y, test_set):
        """Check if test_set d-separates x and y.

        Parameters
        ----------
        x : set of str
        
        y : set of str
        
        test_set : set of str

        Returns
        ----------
        Bool
            If test_set d-separates x and y, return True else return False.
        """
        return nx.d_separated(self.explicit_unob_var_dag, x, y, test_set)

    @property
    def c_components(self):
        """Return the C-component set of the graph.

        Returns
        ----------
        set of str
            The C-component set of the graph
        """
        bi_directed_graph = nx.Graph()
        bi_directed_graph.add_nodes_from(self.dag.nodes)
        bi_directed_graph.add_edges_from(self.latent_confounding_arcs)
        return nx.connected_components(bi_directed_graph)

    def ancestors(self, x):
        """Return the ancestors of all nodes in x.

        Parameters
        ----------
        x : set of str
            a set of nodes in the graph

        Returns
        ----------
        set of str
            Ancestors of nodes in x in the graph
        """
        g = self.observed_dag

        return ancestors_of_iter(g, x)
    
    def descendents(self, x):
        """Return the descendents of all nodes in x.

        Parameters
        ----------
        x : set of str
            a set of nodes in the graph

        Returns
        ----------
        set of str
            Descendents of nodes x of the graph
        """
        # des = set()
        # x = {x} if isinstance(x, str) else x
        
        # for node in x:
        #     des.add(node)
        #     try:
        #         des.update(nx.descendants(self.observed_dag, node))
        #     except Exception:
        #         pass
        g = self.observed_dag
        
        return descendents_of_iter(g, x)

    def parents(self, x, only_observed=True):
        """Return the direct parents of the node x in the graph.

        Parameters
        ----------
        x : str 
            Name of the node x.
        
        only_observed : bool, optional
            If True, then only find the observed parents in the causal graph,
            otherwise also include the unobserved variables, by default True

        Returns
        -------
        list
            Parents of the node x in the graph
        """
        if only_observed:
            return self.causation[x]
        else:
            return list(self.explicit_unob_var_dag.predecessors(x))

    @property
    def observed_dag(self):
        """Return the observed part of the graph, including observed nodes and
        edges between them.

        Returns
        ----------
        nx.MultiDiGraph
            The observed part of the graph
        """
        edges = []
        for k, v in self.causation.items():
            for para in v:
                edges.append((para, k, 0))
        ob_dag = nx.MultiDiGraph()
        ob_dag.add_edges_from(edges)
        return ob_dag

    @property
    def explicit_unob_var_dag(self):
        """Build a new dag where all unobserved confounding arcs are replaced
        by explicit unobserved variables

        Returns
        ----------
        nx.MultiDiGraph
        """
        new_dag = self.observed_dag
        for i, (node1, node2) in enumerate(self.latent_confounding_arcs):
            new_dag.add_edges_from(
                [(f'U{i}', node1, 'n'), (f'U{i}', node2, 'n')]
            )
        return new_dag

    @property
    def topo_order(self):
        """Retrun the topological order of the nodes in the observed graph

        Returns
        ----------
        generator
            Nodes in the topological order
        """
        return nx.topological_sort(self.observed_dag)

    def add_nodes(self, nodes, new=False):
        """
        If not new, add all nodes in the nodes to the current
        CausalGraph, else create a new graph and add nodes.

        Parameters
        ----------
        nodes : set or list
        
        new : bool, optional
            If new create and return a new graph. Defaults to False.

        Returns
        ----------
        CausalGraph
        """
        ori_nodes = self.dag.nodes

        if not new:
            self.dag.add_nodes_from(nodes)
            for node in nodes:
                if node not in ori_nodes:
                    self.causation[node] = []
        else:
            new_dag = deepcopy(self.dag)
            new_causation = deepcopy(self.causation)
            new_dag.add_nodes_from(nodes)
            for node in nodes:
                if node not in ori_nodes:
                    new_causation[node] = []
            return CausalGraph(new_causation, dag=new_dag)

    def add_edges_from(self, edge_list, new=False, observed=True):
        """
        Add edges to the causal graph.

        Parameters
        ----------
        edge_list : list
            Every element of the list contains two elements, the first for
            the parent
        
        new : bool
            Return a new graph if set as True
        
        observed : bool
            Add unobserved bidirected confounding arcs if not observed.
        """
        if not new:
            if observed:
                for edge in edge_list:
                    self.causation[edge[1]].append(edge[0])
                    self.dag.add_edge(edge[0], edge[1], 0)
            else:
                for edge in edge_list:
                    self.dag.add_edges_from(
                        [(edge[0], edge[1], 'n'), (edge[1], edge[0], 'n')]
                    )
        else:
            new_dag = deepcopy(self.dag)
            new_causation = deepcopy(self.causation)
            if observed:
                new_dag.add_edges_from(edge_list)
                for edge in edge_list:
                    new_causation[edge[1]].append(edge[0])
            else:
                for edge in edge_list:
                    new_dag.add_edges_from(
                        [(edge[0], edge[1], 'n'), (edge[1], edge[0], 'n')]
                    )
            return CausalGraph(new_causation, dag=new_dag)

    def add_edge(self, s, t, observed=True):
        """
        Add an edge between nodes i and j. Add an unobserved latent confounding
        arc if not observed.

        Parameters
        ----------
        s : str
            Source of the edge.
        
        t : str
            Target of the edge.
        
        observed : bool
            Add an unobserved latent confounding arc if False.
        """
        if observed:
            self.dag.add_edge(s, t, 0)
            self.causation[t].append(s)
        else:
            self.dag.add_edge(s, t, 'n')

    def remove_nodes(self, nodes, new=False):
        """
        Remove all nodes in the graph.

        Parameters
        ----------
        nodes : set or list
        
        new : bool, optional
            If True, create a new graph, remove nodes in that graph and return
            it. Defaults to False.

        Returns
        ---------
        CausalGraph
            Return a CausalGraph if new.
        """

        if not new:
            for node in nodes:
                for k in list(self.causation.keys()):
                    if k == node:
                        del self.causation[node]
                        continue
                    try:
                        self.causation[k].remove(node)
                    except Exception:
                        pass
            self.dag.remove_nodes_from(nodes)
        else:
            new_causation = deepcopy(self.causation)
            new_dag = deepcopy(self.dag)
            new_dag.remove_nodes_from(nodes)

            for node in nodes:
                for k in list(new_causation.keys()):
                    if k == node:
                        del new_causation[node]
                        continue
                    try:
                        new_causation[k].remove(node)
                    except Exception:
                        pass
            return CausalGraph(new_causation, dag=new_dag)

    def remove_edge(self, edge, observed=True):
        """
        Remove the edge in the CausalGraph. If observed, remove the unobserved
        latent confounding arcs.

        Parameters
        ----------
        edge : tuple
            2 elements denote the start and end of the edge, respectively        
        
        observed : bool
            If not observed, remove the unobserved latent confounding arcs.
        """
        if observed:
            self.dag.remove_edge(edge[0], edge[1], 0)
            try:
                self.causation[edge[1]].remove(edge[0])
            except Exception:
                pass
        else:
            self.dag.remove_edges_from(
                [(edge[0], edge[1], 'n'), (edge[1], edge[0], 'n')]
            )

    def remove_edges_from(self, edge_list, new=False, observed=True):
        """
        Remove all edges in the edge_list in the graph.

        Parameters
        ----------
        edge_list : list
        
        new : bool, optional
            If new, creat a new CausalGraph and remove edges.
        
        observed : bool, optional
            Remove unobserved latent confounding arcs if not observed.

        Returns
        ----------
        CausalGraph
            Return a new CausalGraph if new.
        """
        if not new:
            if observed:
                for edge in edge_list:
                    self.dag.remove_edge(edge[0], edge[1], 0)
                    try:
                        self.causation[edge[1]].remove(edge[0])
                    except Exception:
                        pass
            else:
                for edge in edge_list:
                    self.dag.remove_edges_from(
                        [(edge[0], edge[1], 'n'), (edge[1], edge[0], 'n')]
                    )
        else:
            new_dag = deepcopy(self.dag)
            new_causation = deepcopy(self.causation)
            if observed:
                for edge in edge_list:
                    new_dag.remove_edge(edge[0], edge[1], 0)
                    try:
                        new_causation[edge[1]].remove(edge[0])
                    except Exception:
                        pass
            else:
                for edge in edge_list:
                    new_dag.remove_edges_from(
                        [(edge[0], edge[1], 'n'), (edge[1], edge[0], 'n')]
                    )
            return CausalGraph(new_causation, new_dag)

    def build_sub_graph(self, subset):
        """Return a new CausalGraph as the subgraph of the graph with nodes in the
        subset.

        Parameters
        ----------
        subset : set

        Returns
        ----------
        CausalGraph
        """
        check_nodes(self.ava_nodes, subset)

        nodes = set(self.causation.keys()).difference(subset)
        return self.remove_nodes(nodes, new=True)

    def remove_incoming_edges(self, x, new=False):
        """Remove incoming edges of all nodes of x. If new, do this in the new
        CausalGraph.

        Parameters
        ----------
        x : set or list
        
        new : bool
            Return a new graph if set as Ture.

        Returns
        ----------
        CausalGraph
            If new, return a subgraph of the graph without all incoming edges
            of nodes in x
        """
        check_nodes(self.ava_nodes, x)

        edges = self.dag.in_edges(x, keys=True)
        o_edges, u_edges = [], []

        for edge in edges:
            if edge[2] == 'n':
                u_edges.append(edge)
            else:
                o_edges.append(edge)

        if new:
            return self.remove_edges_from(o_edges, new).remove_edges_from(
                u_edges, new, observed=False
            )
        else:
            self.remove_edges_from(o_edges, new)
            self.remove_edges_from(u_edges, new, observed=False)

    def remove_outgoing_edges(self, x, new=False):
        """Remove outcoming edges of all nodes in x.

        Parameters
        ----------
        x : set
        
        new : bool

        Returns
        ----------
        CausalGraph
            If new, return a subgraph of the graph without all outcoming edges
            of nodes in x.
        """
        check_nodes(self.ava_nodes, x)

        removing_edges = [
            edge for edge in self.dag.out_edges(x, keys=True) if edge[2] == 0
        ]
        return self.remove_edges_from(removing_edges, new, observed=True)

    def plot(self, **kwargs):
        ng = nx.DiGraph(self.causation)
        options = dict(with_labels=True, node_size=1000, **kwargs)
        nx.draw(ng, **options)

    def __repr__(self):
        return to_repr(self)




from copy import deepcopy
from collections import defaultdict

import numpy as np

from sklearn import clone
from sklearn.preprocessing import OneHotEncoder, OrdinalEncoder
from sklearn.model_selection import KFold


from . import propensity_score
from .base_models import BaseEstModel
from .utils import (convert2array, get_wv, convert4onehot,
                    cartesian, get_tr_ctrl)


class _YModelWrapper:
    def __init__(self, y_model, x_d, y_d):
        self.model = clone(y_model)
        self.x_d = x_d
        self.y_d = y_d

    def fit(self, wvx, target, **kwargs):
        self.model.fit(wvx, target, **kwargs)
        return self

    def predict(self, wvx):
        wv = wvx[:, :-self.x_d]
        n = wv.shape[0]
        y_nji = np.full((n, self.y_d, self.x_d), np.NaN)

        for i in range(self.x_d):
            x = np.zeros((n, self.x_d))
            x[:, i] == 1
            x = np.concatenate((wv, x), axis=1)
            y_nji[:, :, i] = self.model.predict(x).reshape(n, self.y_d)

        return y_nji


class DoublyRobust(BaseEstModel):
    # TODO: consider combined treatment when dealing with multiple treatment case
    r"""
    The doubly robust estimator has 3 steps
    (see [1] for reference, see also the slides [2])

    The procedure is composed of 3 steps.
    1. Let k (cf_folds in our class) be an int. Form a k-fold random
    partition {..., (train_data_i, test_data_i), ...,
    (train_data_k, test_data_k)}.

    2. For each i, train y_model and x_model on train_data_i, then evaluate
    their performances in test_data_i whoes results will be saved as
    (y_hat_k, x_hat_k). Note that x_model should be a machine learning model
    with a predict_proba method. All (y_hat_k, x_hat_k) will be combined to
    give the new dataset (y_hat, x_hat). Note that for each y_hat_k, we must
    evaluate its values on all possible treatments, i.e., y_hat should have 
    the shape (n, y_d, x_d) where n is the number of examples, y_d is the
    dimension of y, and x_d is the dimension of x (the number of possible
    treatment values). Note that if there are multiple treatments, we can
    either use the combined_treatment technique to covert the multiple
    discrete classification tasks into a single discrete classification task
    or simply train multiple models for each classification task. For an
    example, if there are two different binary treatments:
        treatment_1: x_1 | x_1 \in {'sleep', 'run'},
        treatment_2: x_2 | x_2 \in {'study', 'work'},
    then we can convert to these two binary classification tasks into a single
    classification with 4 different classes:
        treatment: x | x \in {0, 1, 2, 3},
    where, for example, 1 stands for ('sleep' and 'stuy').
    TODO: there might be some issuses when considering multi-dim y, but im not
    sure currently.

    3. Build the final dataset (v, y_prime_treat - y_prime_control) where y_prime
    is difined as
        y_prime_treat = y_hat_x + \frac{(y - y_hat_x) * I(X=x)}{x_hat_x}
    and train the final yx_model on this dataset to predict the causal effect, 
        y_hat_treat - y_hat_control,
    on v. Then we can directly estimate the CATE by passing the covariate v to
    the model. Note that we train several independent models for each value of
    treatment, thus if there are i different values for treatment (use the
    combined_treatment technique for multiple treatments) we will have i - 1
    different models.

    Attributes
    ----------
    _is_fitted : bool
        True if the model is fitted ortherwise False.

    x_model : estimator
        Any valid x_model should implement the fit and predict_proba methods

    y_model : estimator
        Any valid y_model should implement the fit and predict methods

    yx_model : estimatro
        Any valid yx_model should implement the fit and predict methods

    cf_fold : int, optional
        The nubmer of folds for performing cross fit, by default 1

    random_state : int, optional
        Random seed, by default 2022

    categories : str, optional
        by default 'auto'

    x_hat_dict : dict
        Cached values when fitting the x_model.

    y_hat_dict : dict
        Cached values when fitting the y_model.

    ord_transformer : OrdinalEncoder
        Ordinal transformer of the discrete treament.

    oh_transformer : OneHotEncoder
        One hot encoder of the discrete treatment. Note that the total transformer
        is combined by the ord_transformer and oh_transformer. See comp_transformer
        for detail.

    label_dict : dict        

    Methods
    ----------
    fit(data, outcome, treatment, adjustment, covariate)
        Fit the DoublyRobust estimator model.

    estimate(data, treat, control, quantity)
        Estimate the causal effect.

    comp_transformer(x, categories='auto')
        Transform the discrete treatment into one-hot vectors.

    _cross_fit(model)
        Fit x_model and y_model in a cross fitting manner.

    _fit_first_stage(x_model, y_model, y, x, wv, folds)
        Fit the first stage of the double machine learning.

    _fit_second_stage(yx_model, y_prime, x_prime)
        Fit the second stage of the DML.

    _prepare4est(data)

    _gen_x_model()

    _gen_y_model

    Reference
    ----------
    [1] https://www.ncbi.nlm.nih.gov/pmc/articles/PMC3070495/pdf/kwq439.pdf
    [2] https://www4.stat.ncsu.edu/~davidian/double.pdf 
    """

    def __init__(
        self,
        x_model,
        y_model,
        yx_model,
        cf_fold=1,
        random_state=2022,
        categories='auto',
    ):
        """
        Parameters
        ----------
        x_model : estimator
            The machine learning model which is trained to modeling the treatment. 
            Any valid x_model should implement the :py:func:`fit()` and :py:func:`predict_proba()` methods

        y_model : estimator
            The machine learning model which is trained to modeling the outcome with covariates (possibly adjustment) and the  treatment. 
            Any valid y_model should implement the fit and predict methods

        yx_model : estimatro
            The machine learning model which is trained in the final stage of doubly robust method to modeling the causal effects with covariates (possibly adjustment). 
            Any valid yx_model should implement the fit and predict methods

        cf_fold : int, optional
            The nubmer of folds for performing cross fit in th first stage, by default 1

        random_state : int, optional
            Random seed, by default 2022

        categories : str, optional
            by default 'auto'
        """
        self.cf_fold = cf_fold
        self.x_model = clone(x_model)
        self.y_model = clone(y_model)
        self.yx_model = clone(yx_model)

        self.x_hat_dict = defaultdict(list)
        self.y_hat_dict = defaultdict(list)
        self.x_hat_dict['is_fitted'].append(False)
        self.y_hat_dict['is_fitted'].append(False)

        super().__init__(
            random_state=random_state,
            categories=categories,
            is_discrete_treatment=True,
        )

    def fit(
        self,
        data,
        outcome,
        treatment,
        treat=None,
        control=None,
        adjustment=None,
        covariate=None,
    ):
        """Fit the DoublyRobust estimator model. Note that the trainig of a doubly robust model has three stages, where we implement them in 
        :py:func:`_fit_1st_stage` and :py:func:`_fit_2nd_stage`.

        Parameters
        ----------
        data : pandas.DataFrame
            The dataset used for training the model

        outcome : str or list of str, optional
            Names of the outcome variables

        treatment : str or list of str
            Names of the treatment variables

        treat : float or ndarray, optional
            In the case of single discrete treatment, treat should be an int or
            str in one of all possible treatment values which indicates the
            value of the intended treatment;
            in the case of multiple discrete treatment, treat should be a list
            or an ndarray where treat[i] indicates the value of the i-th intended
            treatment. For example, when there are multiple
            discrete treatments, array(['run', 'read']) means the treat value of
            the first treatment is taken as 'run' and that of the second treatment
            is taken as 'read'.

        control : float or ndarray, optional
            This is similar to the cases of treat, by default None

        adjustment : str or list of str, optional
            Names of the adjustment variables, by default None

        covariate : str or list of str, optional
            Names of the covariate variables, by default None

        Returns
        -------
        instance of DoublyRobust
            The fitted estimator model.
        """
        assert adjustment is not None or covariate is not None, \
            'Need adjustment or covariate to perform estimation.'
        super().fit(
            data, outcome, treatment,
            adjustment=adjustment,
            covariate=covariate,
        )

        # get numpy data
        n = len(data)
        y, x, w, v = convert2array(
            data, outcome, treatment, adjustment, covariate
        )

        self._v = v
        # get the number of cross fit folds
        cfold = self.cf_fold

        if self.categories == 'auto' or self.categories is None:
            categories = 'auto'
        else:
            categories = list(self.categories)

        # transform x into one_hot vectors
        x = self.comp_transormer(x, categories=categories)
        self._x_d = x.shape[1]
        self._y_d = y.shape[1]

        wv = get_wv(w, v)
        # generate y model
        y_model = self._gen_y_model(self.y_model)

        # step 1
        if cfold > 1:
            cfold = int(cfold)
            folds = [
                KFold(n_splits=cfold).split(x), KFold(n_splits=cfold).split(y)
            ]
        else:
            folds = None

        # step 2: cross fit to give the estimated y and x
        self.x_hat_dict, self.y_hat_dict = self._fit_1st_stage(
            self.x_model, y_model, y, x, wv, folds=folds
        )
        x_hat = self.x_hat_dict['paras'][0].reshape((x.shape))
        y_hat = self.y_hat_dict['paras'][0]  # y_hat has shape (n, j, i)

        # step 3
        # calculate the estimated y
        y_prime = np.full((n, y.shape[1], x.shape[1]), np.NaN)

        for i in range(self._x_d):
            y_nji = y_hat[:, :, i]
            x_hat_i = x_hat[:, i].reshape(-1, 1)
            # print(np.mean(x[:, i] - x_hat[:, i], axis=0))
            y_nji += ((y - y_nji) * (x[:, i] ==
                      1).reshape(-1, 1)) / (x_hat_i + 1e-5)
            y_prime[:, :, i] = y_nji

        self._final_result = self._fit_2nd_stage(
            self.yx_model, v, y_prime, control, treat
        )

        self._is_fitted = True

        return self

    def estimate(
        self,
        data=None,
        quantity=None,
        treat=None,
        all_tr_effects=False,
    ):
        """Estimate the causal effect.

        Parameters
        ----------
        data : pd.DataFrame, optional
            The test data for the estimator to evaluate the causal effect, note
            that the estimator directly evaluate all quantities in the training
            data if data is None, by default None

        quantity : str, optional
            The possible values of quantity include:
                'CATE' : the estimator will evaluate the CATE;
                'ATE' : the estimator will evaluate the ATE;
                None : the estimator will evaluate the ITE or CITE, by default None

        all_tr_effects : bool
            If True, return all treatment effects with different treatments, otherwise
            only return the treatment effect of the treatment with the value of 
            treat if treat is provided. If treat is not provided, then the value of
            treatment is taken as the value of that when fitting
            the estimator model.

        treat : float or ndarray, optional
            In the case of single discrete treatment, treat should be an int or
            str in one of all possible treatment values which indicates the
            value of the intended treatment;
            in the case of multiple discrete treatment, treat should be a list
            or an ndarray where treat[i] indicates the value of the i-th intended
            treatment. For example, when there are multiple
            discrete treatments, array(['run', 'read']) means the treat value of
            the first treatment is taken as 'run' and that of the second treatment
            is taken as 'read'.

        Returns
        -------
        ndarray
            The estimated causal effect with the type of the quantity.
        """
        # shape (n, y_d, x_d)
        y_pred_nji = self._prepare4est(
            data=data, all_tr=all_tr_effects, treat=treat
        )

        if quantity == 'CATE':
            assert self._v is not None, 'Need covariates to estimate CATE.'
            return y_pred_nji.mean(axis=0)
        elif quantity == 'ATE':
            return y_pred_nji.mean(axis=0)
        else:
            return y_pred_nji

    def comp_transormer(self, x, categories='auto'):
        """Transform the discrete treatment into one-hot vectors.

        Parameters
        ----------
        x : ndarray, shape (n, x_d)
            An array containing the information of the treatment variables

        categories : str or list, optional
            by default 'auto'

        Returns
        -------
        ndarray
            The transformed one-hot vectors
        """
        if x.shape[1] > 1:
            # if not self._is_fitted:
            if not hasattr(self, 'ord_transformer'):
                self.ord_transformer = OrdinalEncoder(categories=categories)
                self.ord_transformer.fit(x)

                labels = [
                    np.arange(len(c)) for c in self.ord_transformer.categories_
                ]
                labels = cartesian(labels)
                categories = [np.arange(len(labels))]

                self.label_dict = {tuple(k): i for i, k in enumerate(labels)}

            x_transformed = self.ord_transformer.transform(x).astype(int)
            x = np.full((x.shape[0], 1), np.NaN)

            for i, x_i in enumerate(x_transformed):
                x[i] = self.label_dict[tuple(x_i)]

        # if not self._is_fitted:
        if not hasattr(self, 'oh_transformer'):
            self.oh_transformer = OneHotEncoder(categories=categories)
            self.oh_transformer.fit(x)

        x = self.oh_transformer.transform(x).toarray()

        return x

    def _effect_nji_all(self, data=None):
        y_nji = self._prepare4est(data=data, all_tr=True)
        return y_nji

    def effect_nji(self, data=None):
        y_nji = self._prepare4est(data=data, all_tr=False).reshape(-1, self._y_d, 1)
        null_effect = np.zeros_like(y_nji)
        
        if self.treat > self.control:
            effect_ = np.concatenate((null_effect, y_nji), axis=2)
        else:
            effect_ = np.concatenate((y_nji, null_effect), axis=2)
        
        return effect_

    def _prepare4est(self, data=None, all_tr=False, treat=None):
        if not all((self.x_hat_dict['is_fitted'][0],
                   self.y_hat_dict['is_fitted'][0],
                   self._final_result['is_fitted'][0])):
            raise Exception('The model has not been fully fitted yet.')

        v = self._v if data is None else convert2array(data, self.covariate)[0]

        if v is None:
            y_pred_nji = self._final_result['effect']
        else:
            n = v.shape[0]
            y_pred_nji = np.full((n, self._y_d, self._x_d), np.NaN)

            for i in range(self._x_d):
                model = self._final_result['models'][i]
                y_pred_nji[:, :, i] = model.predict(v).reshape(n, self._y_d)

        if all_tr:
            return y_pred_nji
        else:
            if treat is None:
                return y_pred_nji[:, :, self.treat]
            else:
                treat = get_tr_ctrl(
                    treat,
                    self.comp_transormer,
                    treat=True,
                    one_hot=False,
                    discrete_treat=True,
                )
                return y_pred_nji[:, :, treat]

    def _fit_1st_stage(
        self,
        x_model,
        y_model,
        y, x, wv,
        folds=None,
        **kwargs
    ):
        y = y.squeeze()

        if folds is not None:
            x_folds, y_folds = folds
        else:
            x_folds, y_folds = None, None

        wvx = np.concatenate((wv, x), axis=1)
        # convert back to a vector with each dimension being a value
        # indicating the corresponding discrete value
        label = convert4onehot(x)

        x_hat_dict = self._cross_fit(
            x_model, wv, target=label, folds=x_folds, is_ymodel=False, **kwargs
        )
        y_hat_dict = self._cross_fit(
            y_model, wvx, target=y, folds=y_folds, is_ymodel=True, **kwargs
        )

        return x_hat_dict, y_hat_dict

    def _fit_2nd_stage(
        self,
        yx_model,
        v,
        y_prime,
        control,
        treat,
    ):
        final_result = defaultdict(list)
        final_result['is_fitted'].append(False)

        control = get_tr_ctrl(
            control,
            self.comp_transormer,
            treat=False,
            one_hot=False,
            discrete_treat=True,
        )
        treat = get_tr_ctrl(
            treat,
            self.comp_transormer,
            treat=True,
            one_hot=False,
            discrete_treat=True,
        )
        self.treat = treat
        self.control = control

        y_prime_control = (y_prime[:, :, control]).reshape(-1, self._y_d, 1)
        y_prime = y_prime - y_prime_control

        if v is None:
            final_result['effect'] = y_prime
        else:
            for i in range(self._x_d):
                model = clone(yx_model)
                model.fit(v, y_prime[:, :, i].squeeze())
                final_result['models'].append(model)

        final_result['is_fitted'][0] = True

        return final_result

    def _cross_fit(self, model, *args, **kwargs):
        folds = kwargs.pop('folds', None)
        is_ymodel = kwargs.pop('is_ymodel', False)
        target = kwargs.pop('target')
        fitted_result = defaultdict(list)

        if folds is None:
            wv = args[0]
            model.fit(wv, target, **kwargs)

            if not is_ymodel:
                p_hat = model.predict_proba(wv)
            else:
                p_hat = model.predict(wv)

            fitted_result['models'].append(deepcopy(model))
            fitted_result['paras'].append(p_hat)
            idx = np.arange(start=0, stop=wv.shape[0])
            fitted_result['train_test_id'].append((idx, idx))
        else:
            for i, (train_id, test_id) in enumerate(folds):
                model_ = deepcopy(model)
                temp_wv = args[0][train_id]
                temp_wv_test = args[0][test_id]
                target_train = target[train_id]
                model_.fit(temp_wv, target_train, **kwargs)
                n, y_d, x_d = target.shape[0], self._y_d, self._x_d

                if not is_ymodel:
                    target_predict = model_.predict_proba(temp_wv_test)
                    if i == 0:
                        target_required = np.full((n, x_d), np.NaN)
                        fitted_result['paras'].append(target_required)
                else:
                    target_predict = model_.predict(temp_wv_test)
                    if i == 0:
                        target_required = np.full((n, y_d, x_d), np.NaN)
                        fitted_result['paras'].append(target_required)

                fitted_result['models'].append(model_)
                fitted_result['paras'][0][test_id] = target_predict
                fitted_result['train_test_id'].append((train_id, test_id))

        fitted_result['is_fitted'] = [True]

        return fitted_result

    def _gen_x_model(self):
        # if use x_model without methods like fit, predict_proba...
        # may have to define new wrappers
        # TODO
        pass

    def _gen_y_model(self, model):
        """Generate a model for fitting the outcome model for the doubly robust
        model.

        Parameters
        ----------
        model : estimator
            Any valid model should implement the fit() and predict() methods.

        Returns
        -------
        instance of _YModelWrapper
            A wrapped model.
        """
        y_model = clone(model)
        return _YModelWrapper(y_model=y_model, x_d=self._x_d, y_d=self._y_d)


class _DoublyRobustOld(BaseEstModel):
    r"""
    The doubly robust estimator has 3 steps
    (see https://www.ncbi.nlm.nih.gov/pmc/articles/PMC3070495/pdf/kwq439.pdf
    for reference, see also the slides
    https://www4.stat.ncsu.edu/~davidian/double.pdf):
    1. Split the data into the treatment group (xt_data) and control group
        (x0_data),then fit two models in these groups to predict the outcome y:
       yt_i = xt_model.predict(w_i), y0_i = x0_model.predict(w_i)
    2. Estimate the propensity score
       ps(w_i) = ps_model.predict(w_i)
    3. Calculate the final result (expected result, note that ps_model should
        be a multi-classification for discrete treatment):
        1/n \sum_i^n [
            (\frac{I(x_i=xt)y_i}{ps_{x_i=xt}(w_i)}
            - yt_i\frac{I(x_i=xt)-ps_{x_i=xt}(w_i)}{ps_{x_i=xt}(w_i)})
            (\frac{I(x_i=x0)y_i}{ps_{x_i=x0}(w_i)}
            - y0_i\frac{I(x_i=x0)-ps_{x_i=x0}(w_i)}{ps_{x_i=x0}(w_i)})
        ]

    Attributes
    ----------
    ml_model_dic : dict
        A dictionary of default machine learning sklearn models currently
        including
            'LR': LinearRegression
            'LogisticR': LogisticRegression.
    ps_model : PropensityScore
    xt_model : MLModel, optional
        The machine learning model trained in the treated group.
    x0_model : MLModel, optional
        The machine learning model trained in the control group.

    Methods
    ----------
    _prepare4est(data, outcome, treatment, adjustment, individual=None)
        Prepare (fit the model) for estimating various quantities including
        ATE, CATE, ITE, and CITE.
    estimate(data, outcome, treatment, adjustment, quantity='ATE',
                 condition_set=None, condition=None, individual=None)
        Integrate estimations for various quantities into a single method.
    estimate_ate(self, data, outcome, treatment, adjustment)
    estimate_cate(self, data, outcome, treatment, adjustment,
                      condition_set, condition)
    estimate_ite(self, data, outcome, treatment, adjustment, individual)
    estimate_cite(self, data, outcome, treatment, adjustment,
                      condition_set, condition, individual)
    """

    def __init__(self, ps_model='LogisticR', est_model='LR'):
        """
        Parameters
        ----------
        ps_model : PropensityScore or str
        est_model : str, optional
            All valid est_model should implement the fit and predict methods
            which can be done by wrapping the corresponding machine learning
            models with the class MLModel.
        """
        super().__init__()

        if type(ps_model) is str:
            ps_model = self.ml_model_dic[ps_model]
        if type(est_model) is str:
            est_model = self.ml_model_dic[est_model]

        # TODO: should we train different regression models for different
        # treatment groups or simply train one model once to model the
        # relation between treatment and adjustment?
        self.ps_model = propensity_score.PropensityScore(ml_model=ps_model)
        self.xt_model = est_model
        self.x0_model = deepcopy(est_model)

    def _prepare4est(self, data, outcome, treatment, adjustment,
                     individual=None, treatment_value=None):
        # TODO: categorical treatment. Currently I hope to convert categorical
        # treatments to integers such that treatment={1, 2,...,n} where n is
        # the number of different treatments. Is there any better idea?

        # The default treatment group data are those whose data[treatment] == 1
        num_treatment = data[treatment].value_counts().shape[0]
        if treatment_value is None or num_treatment == 2:
            treatment_value = 1

        # step 1, fit the treatment group model and the control group model
        xt_data = data.loc[data[treatment] == treatment_value]
        x0_data = data.loc[data[treatment] == 0]
        self.xt_model.fit(xt_data[adjustment], xt_data[outcome])
        self.x0_model.fit(x0_data[adjustment], x0_data[outcome])

        # step 2, fit the propensity score model
        self.ps_model.fit(data, treatment, adjustment)

        # step 3, calculate the final result
        if individual:
            data_ = individual
        else:
            data_ = data
        yt = self.xt_model.predict(data_[adjustment])
        y0 = self.x0_model.predict(data_[adjustment])
        x, y = data_[treatment], data_[outcome]
        # eps = 1e-7

        x0_index = (x == 0).astype(int)
        x0_prob = self.ps_model.predict_proba(data_, adjustment, 0)
        xt_index = (x == treatment_value).astype(int)
        xt_prob = self.ps_model.predict_proba(
            data_, adjustment, treatment_value
        )
        result = (
            (y * xt_index / xt_prob + (xt_prob - xt_index) * yt / xt_prob)
            - (y * x0_index / x0_prob + (x0_prob - x0_index) * y0 / x0_prob)
        )
        return result



from cProfile import label
import math

import numpy as np
import pandas as pd
import torch
from sklearn.preprocessing import OneHotEncoder
from torch.distributions import Categorical, Independent, MixtureSameFamily, \
    Normal
from torch.utils.data import Dataset


def cartesian(arrays):
    n = len(arrays)
    cart_prod = np.array(np.meshgrid(*arrays)).T.reshape(-1, n)
    return cart_prod


def _get_wv(w, v):
    if w is None:
        wv = v
    else:
        if v is not None:
            wv = np.concatenate((w, v), axis=1)
        else:
            wv = w

    return wv


def get_wv(*wv):
    return np.concatenate([w for w in wv if w is not None], axis=1)


def get_tr_ctrl(tr_crtl, trans, *, treat=False, one_hot=False, discrete_treat=True):
    if tr_crtl is None or not discrete_treat:
        tr_crtl = 1 if treat else 0
    else:
        if not isinstance(tr_crtl, np.ndarray):
            if not isinstance(tr_crtl, (list, tuple)):
                tr_crtl = [tr_crtl]
            tr_crtl = np.array(tr_crtl).reshape(1, -1)

        tr_crtl = trans(tr_crtl).reshape(1, -1)

        if not one_hot:
            tr_crtl = convert4onehot(tr_crtl).astype(int)[0]

    return tr_crtl


def get_treat_control(treat_ctrl, trans, treat=False):
    n_treat = len(trans.categories_)

    if treat_ctrl is not None:
        if not isinstance(treat_ctrl, int):
            assert len(treat_ctrl) == n_treat

        treat_ctrl = np.array(list(treat_ctrl))
        treat_ctrl = trans.transform(treat_ctrl.reshape(1, -1))
    else:
        if treat:
            treat_ctrl = np.ones((1, n_treat)).astype(int)
        else:
            treat_ctrl = np.zeros((1, n_treat)).astype(int)

    return treat_ctrl


def shapes(*tensors, all_dim=False):
    shapes = []
    if all_dim:
        for tensor in tensors:
            if tensor is not None:
                shapes.append(tensor.shape)
    else:
        for tensor in tensors:
            if tensor is not None:
                shapes.append(tensor.shape[1])

    return shapes


def nd_kron(x, y):
    assert x.shape[0] == y.shape[0]
    fn = np.vectorize(np.kron, signature='(n),(m)->(k)')
    kron_prod = fn(x, y)

    return kron_prod


def tensor_or_none(x):
    if x is not None:
        return torch.tensor(x)
    else:
        return None


def convert2tensor(*arrays):
    # arrays = list(arrays)
    # for i, array in enumerate(arrays):
    #     if array is not None:
    #         arrays[i] = torch.tensor(array)
    #
    # return arrays
    return tuple(map(tensor_or_none, arrays))


def convert4onehot(x):
    return np.dot(x, np.arange(0, x.shape[1]).T)


def get_groups(target, a, one_hot, *arrays):
    arrays = list(arrays)
    
    if one_hot:
        a = convert4onehot(a)
        label = (a == target)
    # label = np.all(a == target, axis=1)
    else:
        label = np.all(a == target, axis=1)

    for i, array in enumerate(arrays):
        arrays[i] = array[label]

    return arrays


def convert2array(data, *S, tensor=False):
    assert isinstance(data, pd.DataFrame)

    def _get_array(cols):
        if cols is not None:
            r = data[cols].values
            if len(r.shape) == 1:
                r = np.expand_dims(r, axis=1)
        else:
            r = None
        return r

    S = map(_get_array, S)

    if tensor:
        S = map(tensor_or_none, S)

    return tuple(S)


def convert2str(*S):
    S = list(S)
    for i, s in enumerate(S):
        if isinstance(s, str):
            S[i] = tuple(s)
    return S


def one_hot_transformer(*S):
    transformer_list = []

    for s in S:
        if s[0]:
            temp_transormer = OneHotEncoder()
            temp_transormer.fit(s[1])
        else:
            temp_transormer = None

        transformer_list.append(temp_transormer)

    return transformer_list

#
# class DiscreteIOBatchData(Dataset):
#     def __init__(
#         self,
#         X=None,
#         W=None,
#         y=None,
#         X_test=None,
#         y_test=None,
#         train=True,
#     ):
#         if train:
#             self.w = W
#             self.data = torch.argmax(X, dim=1)
#             self.target = torch.argmax(y, dim=1)
#         else:
#             self.w = W
#             self.data = X_test
#             self.target = y_test
#
#     def __len__(self):
#         return self.target.shape[0]
#
#     def __getitem__(self, index):
#         return self.data[index], self.w[index, :], self.target[index]
#
#
# class DiscreteIBatchData(Dataset):
#     def __init__(
#         self,
#         X=None,
#         W=None,
#         y=None,
#         X_test=None,
#         y_test=None,
#         train=True,
#     ):
#         if train:
#             self.w = W
#             self.data = torch.argmax(X, dim=1)
#             self.target = y
#         else:
#             self.w = W
#             self.data = X_test
#             self.target = y_test
#
#     def __len__(self):
#         return self.target.shape[0]
#
#     def __getitem__(self, index):
#         return self.data[index], self.w[index, :], self.target[index, :]
#
#
# class DiscreteOBatchData(Dataset):
#     def __init__(
#         self,
#         X=None,
#         W=None,
#         y=None,
#         X_test=None,
#         y_test=None,
#         train=True,
#     ):
#         if train:
#             self.w = W
#             self.data = X
#             self.target = torch.argmax(y, dim=1)
#         else:
#             self.w = W
#             self.data = X_test
#             self.target = y_test
#
#     def __len__(self):
#         return self.target.shape[0]
#
#     def __getitem__(self, index):
#         return self.data[index, :], self.w[index, :], self.target[index]


class BatchData(Dataset):
    def __init__(
            self,
            X=None,
            W=None,
            y=None,
            X_test=None,
            y_test=None,
            train=True,
    ):
        if train:
            self.w = W
            self.data = X
            self.target = y
        else:
            self.w = W
            self.data = X_test
            self.target = y_test

    def __len__(self):
        return self.target.shape[0]

    def __getitem__(self, index):
        return self.data[index], self.w[index], self.target[index]


class DiscreteIOBatchData(BatchData):
    def __init__(
            self,
            X=None,
            W=None,
            y=None,
            X_test=None,
            y_test=None,
            train=True,
    ):
        if train:
            X = torch.argmax(X, dim=1)
            y = torch.argmax(y, dim=1)

        super(DiscreteIOBatchData, self).__init__(
            X=X, W=W, y=y, X_test=X_test, y_test=y_test, train=train)


class DiscreteIBatchData(BatchData):
    def __init__(
            self,
            X=None,
            W=None,
            y=None,
            X_test=None,
            y_test=None,
            train=True,
    ):
        if train:
            X = torch.argmax(X, dim=1)

        super(DiscreteIBatchData, self).__init__(
            X=X, W=W, y=y, X_test=X_test, y_test=y_test, train=train)


class DiscreteOBatchData(BatchData):
    def __init__(
            self,
            X=None,
            W=None,
            y=None,
            X_test=None,
            y_test=None,
            train=True,
    ):
        if train:
            y = torch.argmax(y, dim=1)

        super(DiscreteOBatchData, self).__init__(
            X=X, W=W, y=y, X_test=X_test, y_test=y_test, train=train)


class GaussianProb:
    """
    A class for gaussian distribution.

    Attributes
    ----------
    mu : tensor
    sigma : tensor

    Methods
    ----------
    prob(x)
        Return the probability of taking x.
    prod_prob(x)
        Return prob where elements in the last dimension are producted.
    """

    def __init__(self, mu, sigma):
        """
        Parameters
        ----------
        mu : tensor
            Mean of the gaussian distribution with shape
            (b, num_gaussian, out_d), where b is the batch size, num_guassian
            is the number of the mixing components, and out_d is the dimension
            of per gaussian distribution.
        sigma : tensor
            Variance of the gaussian distribution with shape
            (b, num_gaussian, out_d), where b is the batch size, num_guassian
            is the number of the mixing components, and out_d is the dimension
            of per gaussian distribution.
        """
        self.mu = mu
        self.sigma = sigma

    def prob_density(self, y):
        """Return the probability of taking y.

        Parameters
        ----------
        y : tensor
            Shape (b, out_d) where b is the batch size.
        Returns
        ----------
        tensor
            The shape is the same as that of mu.
        """
        y = y.unsqueeze(dim=1).expand_as(self.mu)
        p = (1.0 / math.sqrt(2 * math.pi)) * torch.exp(
            -0.5 * ((y - self.mu) / self.sigma)**2
        ) / self.sigma
        return p

    def mixture_density(self, pi, y, torch_D=False):
        """Can replace the implementation with the one provided by pytorch, see
        (https://pytorch.org/docs/stable/distributions.html#mixturesamefamily)
            for details.
        """
        if torch_D:
            mix = Categorical(pi)
            comp = Independent(Normal(self.mu, self.sigma))
            density = MixtureSameFamily(mix, comp).log_prob(y)
        else:
            p_k = self.prob_density(y)
            pi_k = pi.unsqueeze(dim=2).expand_as(p_k)
            density = torch.sum(p_k * pi_k, dim=1)
        return density

    def prod_prob(self, y):
        """Taking product of the last dimension of returned probability.
        """
        return torch.prod(self.prob(y), dim=2)


# def sample(pi, sigma, mu):
#     """Draw samples from a MoG.
#     """
#     # Choose which gaussian we'll sample from
#     pis = Categorical(pi).sample().view(pi.size(0), 1, 1)
#     # Choose a random sample, one randn for batch X output dims
#     # Do a (output dims)X(batch size) tensor here, so the broadcast works in
#     # the next step, but we have to transpose back.
#     gaussian_noise = torch.randn(
#         (sigma.size(2), sigma.size(0)), requires_grad=False
#     )
#     variance_samples = sigma.gather(1, pis).detach().squeeze()
#     mean_samples = mu.detach().gather(1, pis).squeeze()
#     return (gaussian_noise * variance_samples + mean_samples).transpose(0, 1)
